Neko 1.99.9
A portable framework for high-order spectral element flow simulations
Loading...
Searching...
No Matches
opr_conv1.hip
Go to the documentation of this file.
1/*
2 Copyright (c) 2021-2026, The Neko Authors
3 All rights reserved.
4
5 Redistribution and use in source and binary forms, with or without
6 modification, are permitted provided that the following conditions
7 are met:
8
9 * Redistributions of source code must retain the above copyright
10 notice, this list of conditions and the following disclaimer.
11
12 * Redistributions in binary form must reproduce the above
13 copyright notice, this list of conditions and the following
14 disclaimer in the documentation and/or other materials provided
15 with the distribution.
16
17 * Neither the name of the authors nor the names of its
18 contributors may be used to endorse or promote products derived
19 from this software without specific prior written permission.
20
21 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
22 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
23 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
24 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
25 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
26 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
27 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
28 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
29 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
30 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
31 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
32 POSSIBILITY OF SUCH DAMAGE.
33*/
34
35#include <string.h>
36#include <stdlib.h>
37#include <stdio.h>
38#include <hip/hip_runtime.h>
40#include <device/hip/check.h>
41#include "conv1_kernel.h"
42#include "elem_block_tune.h"
43
44extern "C" {
45 #include <common/neko_log.h>
46}
47
48template < const int >
49int tune_conv1(void *du, void *u,
50 void *vx, void *vy, void *vz,
51 void *dx, void *dy, void *dz,
52 void *drdx, void *dsdx, void *dtdx,
53 void *drdy, void *dsdy, void *dtdy,
54 void *drdz, void *dsdz, void *dtdz,
55 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel,
56 int *ch_sel, int *nwf_sel);
57
58extern "C" {
59
63 void hip_conv1(void *du, void *u,
64 void *vx, void *vy, void *vz,
65 void *dx, void *dy, void *dz,
66 void *drdx, void *dsdx, void *dtdx,
67 void *drdy, void *dsdy, void *dtdy,
68 void *drdz, void *dsdz, void *dtdz,
69 void *jacinv, int *nel, int *gdim, int *lx) {
70
71 static int autotune[17] = { 0 };
72 /* elements per block candidate chosen by the tuner */
73 static int autotune_eb[17] = { 0 };
74 /* chunk candidate chosen for the 1d variant */
75 static int autotune_ch[17] = { 0 };
76 /* wavefronts per block candidate chosen for the mfma variant */
77 static int autotune_nwf[17] = { 0 };
78
79 const dim3 nthrds_1d(1024, 1, 1);
80 const dim3 nthrds_kstep((*lx), (*lx), 1);
81 const dim3 nblcks((*nel), 1, 1);
82
83#define CASE_1D(LX, C) \
84 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
85 conv1_kernel_1d<real, LX, NEKO_CHUNKS(LX, C)> ), \
86 nblcks, NEKO_CHUNKS_NTHRDS(LX, C), 0, \
87 (hipStream_t) glb_cmd_queue, \
88 (real *) du, (real *) u, \
89 (real *) vx, (real *) vy, (real *) vz, \
90 (real *) dx, (real *) dy, (real *) dz, \
91 (real *) drdx, (real *) dsdx, (real *) dtdx, \
92 (real *) drdy, (real *) dsdy, (real *) dtdy, \
93 (real *) drdz, (real *) dsdz, (real *) dtdz, \
94 (real *) jacinv); \
95 HIP_CHECK(hipGetLastError());
96
97/* Runtime dispatch onto the tuned chunk candidate */
98#define CASE_1D_SEL(LX, SEL) \
99 switch (SEL) { \
100 case 0: CASE_1D(LX, 0); break; \
101 case 1: CASE_1D(LX, 1); break; \
102 case 2: CASE_1D(LX, 2); break; \
103 default: CASE_1D(LX, 3); break; \
104 }
105
106#define CASE_KSTEP(LX, C) \
107 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
108 conv1_kernel_kstep<real, LX, NEKO_EB(LX, C)> ), \
109 NEKO_EB_NBLCKS(*nel, LX, C), NEKO_EB_NTHRDS(LX, C), 0, \
110 (hipStream_t) glb_cmd_queue, \
111 (real *) du, (real *) u, \
112 (real *) vx, (real *) vy, (real *) vz, \
113 (real *) dx, (real *) dy, (real *) dz, \
114 (real *) drdx, (real *) dsdx, (real *) dtdx, \
115 (real *) drdy, (real *) dsdy, (real *) dtdy, \
116 (real *) drdz, (real *) dsdz, (real *) dtdz, \
117 (real *) jacinv, *nel); \
118 HIP_CHECK(hipGetLastError());
119
120/* Runtime dispatch onto the tuned candidate */
121#define CASE_KSTEP_SEL(LX, SEL) \
122 switch (SEL) { \
123 case 0: CASE_KSTEP(LX, 0); break; \
124 case 1: CASE_KSTEP(LX, 1); break; \
125 default: CASE_KSTEP(LX, 2); break; \
126 }
127
128#define CASE_MFMA(LX, C) \
129 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
130 conv1_kernel_mfma<real, LX, NEKO_MFMA_NWF(C)> ), \
131 NEKO_MFMA_NBLCKS(*nel, LX, C), NEKO_MFMA_NTHRDS(C), 0, \
132 (hipStream_t) glb_cmd_queue, \
133 (real *) du, (real *) u, \
134 (real *) vx, (real *) vy, (real *) vz, \
135 (real *) dx, (real *) dy, (real *) dz, \
136 (real *) drdx, (real *) dsdx, (real *) dtdx, \
137 (real *) drdy, (real *) dsdy, (real *) dtdy, \
138 (real *) drdz, (real *) dsdz, (real *) dtdz, \
139 (real *) jacinv, *nel); \
140 HIP_CHECK(hipGetLastError());
141
142/* Runtime dispatch onto the tuned wavefronts per block candidate */
143#define CASE_MFMA_SEL(LX, SEL) \
144 switch (SEL) { \
145 case 0: CASE_MFMA(LX, 0); break; \
146 case 1: CASE_MFMA(LX, 1); break; \
147 case 2: CASE_MFMA(LX, 2); break; \
148 case 3: CASE_MFMA(LX, 3); break; \
149 default: CASE_MFMA(LX, 4); break; \
150 }
151
152/* The switch above has to name every wavefront candidate, see the note in
153 ax_helm.hip: a candidate reaching the default arm launches one geometry's
154 grid at another geometry's kernel. */
155static_assert(NEKO_MFMA_NWF_CANDIDATES == 5,
156 "CASE_MFMA_SEL does not cover the candidate space");
157
158#define CASE(LX) \
159 case LX: \
160 if(autotune[LX] == 0 ) { \
161 autotune[LX]=tune_conv1<LX>(du, u, \
162 vx, vy, vz, \
163 dx, dy, dz, \
164 drdx, dsdx, dtdx, \
165 drdy, dsdy, dtdy, \
166 drdz, dsdz, dtdz, \
167 jacinv, nel, gdim, lx, &autotune_eb[LX], \
168 &autotune_ch[LX], \
169 &autotune_nwf[LX]); \
170 } else if (autotune[LX] == 1 ) { \
171 CASE_1D_SEL(LX, autotune_ch[LX]); \
172 } else if (autotune[LX] == 2 ) { \
173 CASE_KSTEP_SEL(LX, autotune_eb[LX]); \
174 } else if (autotune[LX] == 3 ) { \
175 CASE_MFMA_SEL(LX, autotune_nwf[LX]); \
176 } \
177 break
178
179#define CASE_LARGE(LX) \
180 case LX: \
181 CASE_KSTEP(LX, 0); \
182 break
183
184
185 if ((*lx) < 11) {
186 switch(*lx) {
187 CASE(2);
188 CASE(3);
189 CASE(4);
190 CASE(5);
191 CASE(6);
192 CASE(7);
193 CASE(8);
194 CASE(9);
195 CASE(10);
196 default:
197 {
198 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
199 exit(1);
200 }
201 }
202 }
203 else {
204 switch(*lx) {
205 CASE_LARGE(11);
206 CASE_LARGE(12);
207 CASE_LARGE(13);
208 CASE_LARGE(14);
209 CASE_LARGE(15);
210 CASE_LARGE(16);
211 default:
212 {
213 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
214 exit(1);
215 }
216 }
217 }
218 }
219}
220
221template < const int LX >
222int tune_conv1(void *du, void *u,
223 void *vx, void *vy, void *vz,
224 void *dx, void *dy, void *dz,
225 void *drdx, void *dsdx, void *dtdx,
226 void *drdy, void *dsdy, void *dtdy,
227 void *drdz, void *dsdz, void *dtdz,
228 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel,
229 int *ch_sel, int *nwf_sel) {
232 int best1 = 0;
235 int best3 = 0;
236 const int rounds = neko_tune_rounds();
237 const int iters = neko_tune_iters();
238 /* Candidates of the kstep sweep, one -- the unblocked shape -- when the
239 elements per block sweep is off */
240 const int eb_cand = neko_eb_sweep() ? NEKO_EB_CANDIDATES : 1;
241 /* Geometry pinned by each formulation's own variable, -1 to sweep it */
242 const int ch_pin = neko_chunks_pin();
243 const int eb_pin = neko_eb_pin();
244 /* Wavefronts only: the tile dimension is Ax helm's, and these operators
245 stay on the default tile, see neko_mfma_nwf_pin() */
246 const int nwf_pin = neko_mfma_nwf_pin();
247 /* Formulation pinned by NEKO_AUTOTUNE, as the identifier this returns */
248 int strat = 0;
249 /* Hardware capability, which the explicit NEKO_AUTOTUNE=MFMA pin needs */
250 const bool mfma = mfma_lx_supported<LX>() && hip_have_mfma();
251 int best = 0;
252 int retval;
253
254 for (int c = 0; c < NEKO_EB_CANDIDATES; c++) {
256 }
257 for (int c = 0; c < NEKO_CHUNKS_CANDIDATES; c++) {
259 }
260 for (int c = 0; c < NEKO_MFMA_CANDIDATES; c++) {
262 }
263
264 const dim3 nthrds_1d(1024, 1, 1);
265 const dim3 nthrds_kstep((*lx), (*lx), 1);
266 const dim3 nblcks((*nel), 1, 1);
267 const hipStream_t stream = (hipStream_t) glb_cmd_queue;
268
269 char *env_value = NULL;
270 char neko_log_buf[80];
271
272 env_value=getenv("NEKO_AUTOTUNE");
273
274 sprintf(neko_log_buf, "Autotune conv1 (lx: %d)", *lx);
276
277 *eb_sel = 0;
278 *ch_sel = 0;
279 *nwf_sel = 0;
280
281 /*
282 * NEKO_AUTOTUNE names a formulation, and that is all it does: the sweep
283 * below is narrowed to that one kernel family, but its geometry -- the
284 * chunk size, the elements per block, the wavefronts per block -- is still
285 * measured candidate against candidate. A formulation this build or this
286 * device does not have is reported and ignored, leaving the full sweep.
287 */
288 if(env_value) {
289 if( !strcmp(env_value,"1D") ) {
290 strat = 1;
291 } else if( !strcmp(env_value,"KSTEP") ) {
292 strat = 2;
293 } else if( !strcmp(env_value,"MFMA") ) {
294 if (mfma) {
295 strat = 3;
296 } else {
297 sprintf(neko_log_buf, "MFMA strategy not available for this config");
299 }
300 } else {
301 sprintf(neko_log_buf, "Invalid value set for NEKO_AUTOTUNE");
303 }
304 }
305
306 /* Geometry of the pinned formulation, if its own variable fixes that too.
307 Both pinned leaves nothing to measure, so the kernel is launched once and
308 reported, which is what pinning has always done */
309 const int pin = (strat == 1) ? ch_pin : (strat == 2) ? eb_pin :
310 (strat == 3) ? nwf_pin : -1;
311
312 if (pin >= 0) {
313 switch (strat) {
314 case 1:
315 *ch_sel = pin;
317 sprintf(neko_log_buf, "Set by env : 1 (1D, %d chunk)",
319 break;
320 case 2:
321 *eb_sel = pin;
323 sprintf(neko_log_buf, "Set by env : 2 (KSTEP, %d elem/block)",
324 NEKO_EB_SEL(LX, pin));
325 break;
326 default:
327 *nwf_sel = pin;
329 sprintf(neko_log_buf, "Set by env : 3 (MFMA, %d wf, %d elem/block)",
331 break;
332 }
335 return strat;
336 }
337
338 if (strat) {
339 sprintf(neko_log_buf, "Set by env : %d (%s)", strat, env_value);
341 }
342
343 /* Formulations the sweep considers, see NEKO_TUNE_FOR(). NEKO_MFMA_TUNE
344 keeps the matrix core variant out of an unpinned sweep, but naming it
345 explicitly still measures it */
346 const bool try_1d = (strat == 0 || strat == 1);
347 const bool try_kstep = (strat == 0 || strat == 2);
348 const bool try_mfma = mfma && ((strat == 3) ||
349 (strat == 0 && neko_mfma_sweep()));
350
353 /* Warm every variant before timing anything: each specialisation has to be
354 resident and the clocks at steady state, or whichever is timed first is
355 measured on a colder part */
356 for (int i = 0; i < NEKO_TUNE_WARMUP; i++) {
358 CASE_1D_SEL(LX, c);
359 }
361 CASE_KSTEP_SEL(LX, c);
362 }
364 CASE_MFMA_SEL(LX, c);
365 }
366 }
367
368 /* Interleaved rounds, best time per variant */
369 for (int r = 0; r < rounds; r++) {
372 }
375 }
378 }
379 }
380
383
387 *eb_sel = best;
388 *ch_sel = best1;
389 *nwf_sel = best3;
390
391 if (time1[best1] < time2[best]) {
392 retval = 1;
393 } else {
394 retval = 2;
395 }
396
397 /* The mfma variant joins the comparison only where it exists, its
398 candidates are left at NEKO_TUNE_INIT otherwise */
399 if (time3[best3] < ((retval == 1) ? time1[best1] : time2[best])) {
400 retval = 3;
401 }
402
403 /* Leave the chosen kernel's output in place: the tuner stands in for a real
404 evaluation and the variants do not sum in the same order */
405 if (retval == 1) {
407 } else if (retval == 2) {
409 } else {
411 }
412
413 if (retval == 1) {
414 sprintf(neko_log_buf, "Chose : 1 (1D, %d chunk)",
416 } else if (retval == 2) {
417 sprintf(neko_log_buf, "Chose : 2 (KSTEP, %d elem/block)",
419 } else {
420 sprintf(neko_log_buf, "Chose : 3 (MFMA, %d wf, %d elem/block)",
422 }
425 return retval;
426}
__global__ void ale_add_kinematics_kernel(const int n, T *__restrict__ wx, T *__restrict__ wy, T *__restrict__ wz, const T *__restrict__ x_ref, const T *__restrict__ y_ref, const T *__restrict__ z_ref, const T *__restrict__ phi, const T *__restrict__ x, const T *__restrict__ y, const T *__restrict__ z, const kinematics_params_t kin_params)
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ jacinv
const int i
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ u
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdy
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ vz
__global__ void const T *__restrict__ const T *__restrict__ vx
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ vy
#define NEKO_CHUNKS_CANDIDATES
Definition elem_block.h:125
#define NEKO_EB_CANDIDATES
Definition elem_block.h:63
#define NEKO_EB_SEL(LX, SEL)
Definition elem_block.h:108
#define NEKO_CHUNKS_SEL(LX, SEL)
Definition elem_block.h:147
static int neko_chunks_pin()
#define NEKO_TUNE_TIME(T, LAUNCH, LX, C, ITERS)
static int neko_tune_rounds()
static int neko_eb_pin()
#define NEKO_TUNE_LOG(LX, T1, T2)
#define NEKO_TUNE_INIT
#define NEKO_TUNE_BEST(T, BEST, N)
static int neko_tune_iters()
#define NEKO_TUNE_FOR(C, ON, PIN, N)
static int neko_eb_sweep()
#define NEKO_TUNE_WARMUP
#define HIP_CHECK(err)
Definition check.h:8
#define NEKO_TUNE_LOG_MFMA(LX, T3)
#define NEKO_MFMA_NWF_CANDIDATES
#define NEKO_MFMA_CANDIDATES
static bool hip_have_mfma()
static int neko_mfma_nwf_pin()
#define NEKO_MFMA_NWF(C)
#define NEKO_MFMA_EB(LX, C)
static int neko_mfma_sweep()
void log_error(char *msg)
void log_message(char *msg)
void log_end_section()
void log_section(char *msg)
#define CASE_KSTEP_SEL(LX, SEL)
#define CASE(LX)
int tune_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel, int *nwf_sel)
#define CASE_1D_SEL(LX, SEL)
void hip_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx)
Definition opr_conv1.hip:63
#define CASE_MFMA_SEL(LX, SEL)
#define CASE_LARGE(LX)